from pathlib import Path
from unittest import TestCase

from django.test import override_settings
from mock import patch, Mock, MagicMock
from django.conf import settings

from livesync.fswatcher import BaseEventHandler
from livesync.core.handler import LiveReloadRequestHandler


class BaseEventHandlerTestCase(TestCase):
    def setUp(self):
        self.handler = BaseEventHandler()
        self.mock_event = MagicMock(src_path='/path/file', dest_path='/path/newfile')

        self.getmd5_patcher = patch('livesync.fswatcher.handlers.get_md5', return_value='md5hash')
        self.mock_get_md5 = self.getmd5_patcher.start()

        self.handle_patcher = patch.object(self.handler, 'handle')
        self.mock_handle = self.handle_patcher.start()

    def tearDown(self):
        self.getmd5_patcher.stop()
        self.handle_patcher.stop()

    def test_on_deleted_remove_file_from_history(self):
        self.handler.history[self.mock_event.src_path] = "md5hash"
        # act
        self.handler.on_deleted(self.mock_event)
        # assert
        self.assertNotIn(self.mock_event.src_path, self.handler.history)

    def test_on_deleted_invokes_handle(self):
        self.handler.history[self.mock_event.src_path] = "md5hash"
        # act
        self.handler.on_deleted(self.mock_event)
        # assert
        self.mock_handle.assert_called_with(self.mock_event)

    def test_on_moved_updates_history(self):
        self.handler.history[self.mock_event.src_path] = "md5hash"
        # act
        self.handler.on_moved(self.mock_event)
        # assert
        self.assertNotIn(self.mock_event.src_path, self.handler.history)
        self.assertIn(self.mock_event.dest_path, self.handler.history)
        self.assertEqual(self.handler.history[self.mock_event.dest_path], 'md5hash')

    def test_on_moved_adds_to_history(self):
        # act
        self.handler.on_moved(self.mock_event)
        # assert
        self.assertIn(self.mock_event.dest_path, self.handler.history)
        self.assertEqual(self.handler.history[self.mock_event.dest_path], 'md5hash')

    def test_on_moved_invokes_handle(self):
        self.handler.history[self.mock_event.src_path] = "md5hash"
        # act
        self.handler.on_moved(self.mock_event)
        # assert
        self.mock_handle.assert_called_with(self.mock_event)

    def test_on_modified_adds_to_history(self):
        # act
        self.handler.on_modified(self.mock_event)
        # assert
        self.assertIn(self.mock_event.src_path, self.handler.history)
        self.assertEqual(self.handler.history[self.mock_event.src_path], 'md5hash')

    def test_on_modified_invoked_handle(self):
        # act
        self.handler.on_modified(self.mock_event)
        # assert
        self.mock_handle.assert_called_with(self.mock_event)

    def test_on_modified_ignores_unchanged_files(self):
        self.handler.history[self.mock_event.src_path] = "md5hash"
        #act
        self.handler.on_modified(self.mock_event)
        # assert
        self.mock_handle.assert_not_called()



class LiveReloadEventHandlerTestCase(TestCase):
    def setUp(self):
        self.handler = LiveReloadRequestHandler()

    @patch('livesync.core.handler.livesync_event.send')
    def test_handle_dispatches_event(self, mock_dispatch):
        # act
        self.handler.handle(Mock())
        # assert
        mock_dispatch.assert_called()

    @patch('livesync.core.middleware.settings.STATICFILES_DIRS',['foo', ])
    def test_handler_watches_staticfiles_dirs(self):
        self.assertIn('foo', self.handler.watched_paths)


    @patch('livesync.core.middleware.settings.TEMPLATES',[{'DIRS': ['foo', ]}])
    def test_handler_watches_template_dirs(self):
        self.assertIn('foo', self.handler.watched_paths)

    @patch('livesync.core.middleware.settings.STATICFILES_DIRS',['foo', ])
    @patch('livesync.core.middleware.settings.TEMPLATES',[{'DIRS': ['bar', ]}])
    def test_handler_combines_template_dirs_and_static_files(self):
        self.assertIn('foo', self.handler.watched_paths)
        self.assertIn('bar', self.handler.watched_paths)

    @override_settings(BASE_DIR=Path('/tmp/'))
    @patch('livesync.core.handler.get_app_template_dirs')
    def test_handle_pathlib_path(self, mock_app_template_dirs):
        mock_app_template_dirs.return_value = [Path('/tmp')]
        retval = self.handler.watched_paths
        self.assertTrue(hasattr(retval, '__iter__'))



from itertools import zip_longest


def patch_dom_with_vdom(dom_element, virtual_element, level=0, debug=False):
    level += 1

    if dom_element.nodeType == 3:  # text node
        if virtual_element.nodeType == 3:
            if dom_element.textContent != virtual_element.textContent:
                dom_element.textContent = virtual_element.textContent
        else:
            dom_element.replaceWith(virtual_element)
    else:
        if level < 2:  # root element is ignored
            if debug:
                print('ignored', dom_element.nodeName, virtual_element.nodeName)
        else:
            if debug:
                print(dom_element.nodeName, virtual_element.nodeName)

            if dom_element.nodeName != virtual_element.nodeName:
                dom_element.replaceWith(virtual_element)
                return

            for attr in dom_element.attributes:
                if not virtual_element.hasAttribute(attr.name):
                    dom_element.removeAttribute(attr.name)

            for attr in virtual_element.attributes:
                dom_element.setAttribute(attr.name, attr.value)

            if hasattr(virtual_element, 'value'):  # input actual value
                dom_element.value = virtual_element.value

        for old_child, new_child in zip_longest(dom_element.childNodes, virtual_element.childNodes):
            if new_child is None:
                old_child.remove()
            elif old_child is None:
                dom_element.appendChild(new_child)
            else:
                patch_dom_with_vdom(old_child, new_child, level=level)



from io import BytesIO
from typing import TYPE_CHECKING
from xml import sax

from .. import constants

if TYPE_CHECKING:
    from ..environment import Environment


class XmlParser(sax.ContentHandler):
    def __init__(self, environment: 'Environment'):
        self.environment = environment

        self.template = None
        self.act_parent = None

    def parse(self, source: bytes, template_name: str):
        self.template = self.environment.template_class()
        self.act_parent = None

        parser = sax.make_parser()
        #parser.setFeature(sax.handler.feature_external_pes, False)
        #parser.setFeature(sax.handler.feature_external_ges, False)
        parser.setFeature(sax.handler.feature_namespaces, True)
        #parser.setProperty(sax.handler.property_lexical_handler, self)
        parser.setContentHandler(self)

        isource = sax.xmlreader.InputSource()
        isource.setByteStream(BytesIO(source))
        isource.setSystemId(template_name)

        parser.parse(isource)

        return self.template

    def startElementNS(self, name, qname, attrs):
        ns, name = name

        if ns:
            fqdn = '{' + ns + '}' + name
        else:
            fqdn = name

        if (constants.XML_NAMESPACE_FLOW_CONTROL, 'include',) in attrs.keys():  # include another template
            template = self.environment.get_template(attrs[(constants.XML_NAMESPACE_FLOW_CONTROL, 'include')])
            assert template.root_element.xml_tag == name
            element = template.root_element
            self.act_parent.add_child(element)
            self.act_parent = element
        else:
            klass = self.environment.registry.get_class_by_name(fqdn)
            element = klass(xml_tag=name, xml_attrs=attrs)
            if self.act_parent is None:
                assert self.template.root_element is None
                self.template.root_element = self.act_parent = element
            else:
                self.act_parent.add_child(element)
                self.act_parent = element

    def endElementNS(self, *args, **kwargs):
        self.act_parent = self.act_parent.parent

    def characters(self, content):
        if len(content.strip()) == 0:
            return

        content = sax.saxutils.escape(content).strip()
        element = self.environment.registry.anonymus_element_klass(text=content)
        self.act_parent.add_child(element)



from django.urls import path

from . import views

urlpatterns = [
    path('index/', views.index, name='index'),
    path('form/', views.form, name='form'),
    path('data/', views.data, name='data'),
    path('modal/', views.modal, name='modal'),
    path('table/', views.table, name='table'),
    path('panel/', views.panel, name='panel'),
    path('panel/data/', views.panel_data, name='panel-data'),
    path('vfs/poll/', views.poll_filesystem_changes, name='vfs-poll'),
]

app_name = 'backend'



from django.conf import settings
from django.urls import reverse


class DjangoSamonBinding:
    def url(self, view_name, *args, **kwargs):
        return reverse(view_name, args=args, kwargs=kwargs)


def djsamon(request):
    return {
        'djsamon': DjangoSamonBinding()
    }



from browser import document

from kirui.vdom import minify


class MinifyTest:
    def test_minify_text_content(self):
        el = document.createElement('DIV')
        el.innerHTML = """
            <span style="font-weight: bold">
                text
            </span>
            sdfsdfsdf
        """

        assert el.innerHTML != """<span style="font-weight: bold">text </span>sdfsdfsdf  """
        minify(el)
        assert el.innerHTML.strip() == """<span style="font-weight: bold">text </span>sdfsdfsdf"""



#!/usr/bin/env python
import os

from setuptools import find_packages, setup

try:
    with open('VERSION', 'r') as f:
        version = f.read().strip()
except FileNotFoundError:
    version = '0.1.2'

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='kirui',
    version=version,
    description='UI framework with React / Preact and Django SSR backend',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Bence Lovas',
    author_email='me@lovasb.com',
    url='https://kirui.lovasb.com/',
    packages=find_packages(),
    include_package_data=True,
    project_urls={
        'Documentation': 'https://kirui.lovasb.com/',
        'Source': 'https://gitlab.com/lovasb/kirui',
        'Tracker': 'https://gitlab.com/lovasb/kirui/-/issues',
    },
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    scripts=['kirui_dev']
)



import datetime

from django.core.paginator import Paginator, Page, EmptyPage
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.core.cache import cache
from django.template.loader import get_template
from django.views.decorators.csrf import csrf_exempt

from kirui_devserver.backend.forms import SampleForm


def index(request):
    form = SampleForm(request.POST or None)
    form.is_valid()
    return render(request, 'base.xml', context={'now': datetime.datetime.now(), 'form': form})


def form(request):
    form = SampleForm(request.POST or None)
    print(request.POST)
    if form.is_valid():
        form = SampleForm(None)
    else:
        data = get_template('form.xml').render(context={'form': form}, request=request)
        resp = JsonResponse(data, safe=False)
        resp.status_code = 403
        return resp

    return render(request, 'form.xml', context={'now': datetime.datetime.now(), 'form': form})

def tests(request):
    return render(request, 'tests.html')


def poll_filesystem_changes(request):
    return HttpResponse(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))


def data(request):
    return render(request, 'index.xml', context={'iterable': [1, 2, 3], 'value': 8})


def modal(request):
    form = SampleForm(request.POST or None)
    form.now = datetime.datetime.now()
    if form.is_valid():
        pass

    data = get_template('modal.xml').render(context={'form': form}, request=request)
    resp = JsonResponse(data, safe=False)
    if form.is_valid():
        resp.status_code = 200
    else:
        resp.status_code = 403

    return resp


def table(request):
    data = []
    for row in range(1, 20):
        data.append({'first': f'{row}.1', 'second': f'{row}.2'})
    return render(request, 'table.xml', context={'data': data})


def panel(request):
    form = SampleForm(request.POST or None)
    if form.is_valid():
        pass

    return render(request, 'panel.xml', context={'now': datetime.datetime.now(), 'form': form, 'page': None})


TABLE_ROWS = [{'first': f'{_}.1', 'second': f'{_}.2'} for _ in range(0, 100)]


class CustomPage(Page):
    def next_page_number(self):
        try:
            return self.paginator.validate_number(self.number + 1)
        except EmptyPage:
            return ''

    def previous_page_number(self):
        try:
            return self.paginator.validate_number(self.number - 1)
        except EmptyPage:
            return ''


class CustomPaginator(Paginator):
    def _get_page(self, *args, **kwargs):
        return CustomPage(*args, **kwargs)


def panel_data(request):
    form = SampleForm(request.POST or None)
    if form.is_valid():
        TABLE_ROWS = [{'first': f'{_}.1 {form.cleaned_data["masodik"]}', 'second': f'{_}.2'} for _ in range(0, 100)]
        paginate_to = request.POST.get('paginate_to')
        paginate_to = paginate_to or 1
        paginate_to = int(paginate_to)
        p = CustomPaginator(TABLE_ROWS, 20)
        page = p.page(paginate_to)
    else:
        page = None

    data = get_template('panel_data.xml').render(context={'now': datetime.datetime.now(), 'form': form, 'page': page}, request=request)
    resp = JsonResponse(data, safe=False)
    return resp
